
import logging
import os
import os.path
import shutil
import sys
import tempfile
import unittest

import pytest

import fiona
from fiona.collection import supported_drivers
from fiona.errors import FionaValueError, DriverError, SchemaError, CRSError
from fiona.ogrext import calc_gdal_version_num, get_gdal_version_num


logging.basicConfig(stream=sys.stderr, level=logging.INFO)


class ReadingTest(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    @pytest.mark.skipif(not os.path.exists('tests/data/coutwildrnp.gpkg'),
                        reason="Requires geopackage fixture")
    def test_gpkg(self):
        if get_gdal_version_num() < calc_gdal_version_num(1, 11, 0):
            self.assertRaises(DriverError, fiona.open, 'tests/data/coutwildrnp.gpkg', 'r', driver="GPKG")
        else:
            with fiona.open('tests/data/coutwildrnp.gpkg', 'r', driver="GPKG") as c:
                self.assertEquals(len(c), 48)


class WritingTest(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    @pytest.mark.skipif(not os.path.exists('tests/data/coutwildrnp.gpkg'),
                        reason="Requires geopackage fixture")
    def test_gpkg(self):
        schema = {'geometry': 'Point',
                  'properties': [('title', 'str')]}
        crs = {
            'a': 6370997,
            'lon_0': -100,
            'y_0': 0,
            'no_defs': True,
            'proj': 'laea',
            'x_0': 0,
            'units': 'm',
            'b': 6370997,
            'lat_0': 45}

        path = os.path.join(self.tempdir, 'foo.gpkg')
        
        if get_gdal_version_num() < calc_gdal_version_num(1, 11, 0):
            self.assertRaises(DriverError,
                        fiona.open,
                        path,
                        'w',
                        driver='GPKG',
                        schema=schema,
                        crs=crs)
        else:
            with fiona.open(path, 'w',
                            driver='GPKG',
                            schema=schema,
                            crs=crs) as c:
                c.writerecords([{
                    'geometry': {'type': 'Point', 'coordinates': [0.0, 0.0]},
                    'properties': {'title': 'One'}}])
                c.writerecords([{
                    'geometry': {'type': 'Point', 'coordinates': [2.0, 3.0]},
                    'properties': {'title': 'Two'}}])
            with fiona.open(path) as c:
                self.assertEquals(c.schema['geometry'], 'Point')
                self.assertEquals(len(c), 2)



"""Unittests for `$ fio ls`"""


import json
import shutil
import sys
import tempfile
import unittest

from click.testing import CliRunner

import fiona
from fiona.fio.main import main_group

FIXME_WINDOWS = sys.platform.startswith('win')

@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_fio_ls_single_layer():

    result = CliRunner().invoke(main_group, [
        'ls',
        'tests/data/'])
    assert result.exit_code == 0
    assert len(result.output.splitlines()) == 1
    assert json.loads(result.output) == ['coutwildrnp']


@unittest.skipIf(FIXME_WINDOWS, 
                 reason="FIXME on Windows. Please look into why this test is not working.")
def test_fio_ls_indent():

    result = CliRunner().invoke(main_group, [
        'ls',
        '--indent', '4',
        'tests/data/coutwildrnp.shp'])
    assert result.exit_code == 0
    assert len(result.output.strip().splitlines()) == 3
    assert json.loads(result.output) == ['coutwildrnp']


def test_fio_ls_multi_layer():

    infile = 'tests/data/coutwildrnp.shp'
    outdir = tempfile.mkdtemp()
    try:
        
        # Copy test shapefile into new directory
        # Shapefile driver treats a directory of shapefiles as a single
        # multi-layer datasource
        layer_names = ['l1', 'l2']
        for layer in layer_names:
            with fiona.open(infile) as src, \
                    fiona.open(outdir, 'w', layer=layer, **src.meta) as dst:
                for feat in src:
                    dst.write(feat)

        # Run CLI test
        result = CliRunner().invoke(main_group, [
            'ls', outdir])
        assert result.exit_code == 0
        assert json.loads(result.output) == layer_names

    finally:
        shutil.rmtree(outdir)



"""$ fio env"""


import logging

import click

import fiona


@click.command(short_help="Print information about the fio environment.")
@click.option('--formats', 'key', flag_value='formats', default=True,
              help="Enumerate the available formats.")
@click.pass_context
def env(ctx, key):

    """Print information about the Fiona environment: available
    formats, etc.
    """

    verbosity = (ctx.obj and ctx.obj.get('verbosity')) or 1
    logger = logging.getLogger('fio')
    stdout = click.get_text_stream('stdout')
    with fiona.drivers(CPL_DEBUG=(verbosity > 2)) as env:
        if key == 'formats':
            for k, v in sorted(fiona.supported_drivers.items()):
                modes = ', '.join("'" + m + "'" for m in v)
                stdout.write("%s (modes %s)\n" % (k, modes))
            stdout.write('\n')



import logging
import sys

import pytest

import fiona
from fiona.errors import FionaValueError


logging.basicConfig(stream=sys.stderr, level=logging.INFO)

def test_read_fail():
    with pytest.raises(FionaValueError):
        fiona.open('tests/data/coutwildrnp.shp', driver='GeoJSON')
    with pytest.raises(FionaValueError):
        fiona.open('tests/data/coutwildrnp.shp', enabled_drivers=['GeoJSON'])


def test_read():
    with fiona.open(
            'tests/data/coutwildrnp.shp', driver='ESRI Shapefile') as src:
        assert src.driver == 'ESRI Shapefile'
    with fiona.open(
            'tests/data/coutwildrnp.shp',
            enabled_drivers=['GeoJSON', 'ESRI Shapefile']) as src:
        assert src.driver == 'ESRI Shapefile'



"""
Helper objects needed by multiple CLI commands.
"""
from functools import partial
import json
import math
import warnings

from munch import munchify


warnings.simplefilter('default')


def obj_gen(lines):
    """Return a generator of JSON objects loaded from ``lines``."""
    first_line = next(lines)
    if first_line.startswith(u'\x1e'):
        def gen():
            buffer = first_line.strip(u'\x1e')
            for line in lines:
                if line.startswith(u'\x1e'):
                    if buffer:
                        yield json.loads(buffer)
                    buffer = line.strip(u'\x1e')
                else:
                    buffer += line
            else:
                yield json.loads(buffer)
    else:
        def gen():
            yield json.loads(first_line)
            for line in lines:
                yield json.loads(line)
    return gen()


def nullable(val, cast):
    if val is None:
        return None
    else:
        return cast(val)


def eval_feature_expression(feature, expression):
    safe_dict = {'f': munchify(feature)}
    safe_dict.update({
        'sum': sum,
        'pow': pow,
        'min': min,
        'max': max,
        'math': math,
        'bool': bool,
        'int': partial(nullable, int),
        'str': partial(nullable, str),
        'float': partial(nullable, float),
        'len': partial(nullable, len),
    })
    try:
        from shapely.geometry import shape
        safe_dict['shape'] = shape
    except ImportError:
        pass
    return eval(expression, {"__builtins__": None}, safe_dict)


def make_ld_context(context_items):
    """Returns a JSON-LD Context object.

    See http://json-ld.org/spec/latest/json-ld."""
    ctx = {
      "@context": {
        "geojson": "http://ld.geojson.org/vocab#",
        "Feature": "geojson:Feature",
        "FeatureCollection": "geojson:FeatureCollection",
        "GeometryCollection": "geojson:GeometryCollection",
        "LineString": "geojson:LineString",
        "MultiLineString": "geojson:MultiLineString",
        "MultiPoint": "geojson:MultiPoint",
        "MultiPolygon": "geojson:MultiPolygon",
        "Point": "geojson:Point",
        "Polygon": "geojson:Polygon",
        "bbox": {
          "@container": "@list",
          "@id": "geojson:bbox"
        },
        "coordinates": "geojson:coordinates",
        "datetime": "http://www.w3.org/2006/time#inXSDDateTime",
        "description": "http://purl.org/dc/terms/description",
        "features": {
          "@container": "@set",
          "@id": "geojson:features"
        },
        "geometry": "geojson:geometry",
        "id": "@id",
        "properties": "geojson:properties",
        "start": "http://www.w3.org/2006/time#hasBeginning",
        "stop": "http://www.w3.org/2006/time#hasEnding",
        "title": "http://purl.org/dc/terms/title",
        "type": "@type",
        "when": "geojson:when"
      }
    }
    for item in context_items or []:
        t, uri = item.split("=")
        ctx[t.strip()] = uri.strip()
    return ctx


def id_record(rec):
    """Converts a record's id to a blank node id and returns the record."""
    rec['id'] = '_:f%s' % rec['id']
    return rec



import logging
import os
import shutil
import sys
import tempfile
import unittest

import fiona

logging.basicConfig(stream=sys.stderr, level=logging.INFO)

from .test_collection import ReadingTest

def test_index_selection():
    with fiona.open('tests/data/coutwildrnp.shp', 'r', layer=0) as c:
        assert len(c) == 67

class FileReadingTest(ReadingTest):
    
    def setUp(self):
        self.c = fiona.open('tests/data/coutwildrnp.shp', 'r', layer='coutwildrnp')
    
    def tearDown(self):
        self.c.close()

    def test_open_repr(self):
        self.assertEqual(
            repr(self.c),
            ("<open Collection 'tests/data/coutwildrnp.shp:coutwildrnp', mode 'r' "
            "at %s>" % hex(id(self.c))))

    def test_closed_repr(self):
        self.c.close()
        self.assertEqual(
            repr(self.c),
            ("<closed Collection 'tests/data/coutwildrnp.shp:coutwildrnp', mode 'r' "
            "at %s>" % hex(id(self.c))))

    def test_name(self):
        self.assertEqual(self.c.name, 'coutwildrnp')

class DirReadingTest(ReadingTest):
    
    def setUp(self):
        self.c = fiona.open("tests/data", "r", layer="coutwildrnp")
    
    def tearDown(self):
        self.c.close()

    def test_open_repr(self):
        self.assertEqual(
            repr(self.c),
            ("<open Collection 'tests/data:coutwildrnp', mode 'r' "
            "at %s>" % hex(id(self.c))))

    def test_closed_repr(self):
        self.c.close()
        self.assertEqual(
            repr(self.c),
            ("<closed Collection 'tests/data:coutwildrnp', mode 'r' "
            "at %s>" % hex(id(self.c))))

    def test_name(self):
        self.assertEqual(self.c.name, 'coutwildrnp')

    def test_path(self):
        self.assertEqual(self.c.path, "tests/data")

class InvalidLayerTest(unittest.TestCase):

    def test_invalid(self):
        self.assertRaises(ValueError, fiona.open, ("tests/data/coutwildrnp.shp"), layer="foo")

    def test_write_numeric_layer(self):
        self.assertRaises(ValueError, fiona.open,
                          (os.path.join(tempfile.gettempdir(), "test-no-iter.shp")),
                          mode='w', layer=0)



import logging
import sys
import os

import tempfile
import pytest

import fiona


logging.basicConfig(stream=sys.stderr, level=logging.INFO)


def create_sample_data(filename, driver):
    meta = {
        'driver': driver,
        'schema': {
            'geometry': 'Point',
            'properties': {}
        }
    }
    with fiona.open(filename, 'w', **meta) as dst:
        dst.write({
            'geometry': {
                'type': 'Point',
                'coordinates': (0, 0),
            },
            'properties': {},
        })
    assert(os.path.exists(filename))


def test_remove(tmpdir=None):
    if tmpdir is None:
        tmpdir = tempfile.mkdtemp()
    filename_shp = os.path.join(tmpdir, 'test.shp')
    
    create_sample_data(filename_shp, driver='ESRI Shapefile')
    fiona.remove(filename_shp, driver='ESRI Shapefile')
    assert(not os.path.exists(filename_shp))
    
    with pytest.raises(RuntimeError):
        fiona.remove(filename_shp, driver='ESRI Shapefile')

def test_remove_driver(tmpdir=None):
    if tmpdir is None:
        tmpdir = tempfile.mkdtemp()
    filename_shp = os.path.join(tmpdir, 'test.shp')
    filename_json = os.path.join(tmpdir, 'test.json')
        
    create_sample_data(filename_shp, driver='ESRI Shapefile')
    create_sample_data(filename_json, driver='GeoJSON')
    fiona.remove(filename_json, driver='GeoJSON')
    assert(not os.path.exists(filename_json))
    assert(os.path.exists(filename_shp))

def test_remove_collection(tmpdir=None):
    if tmpdir is None:
        tmpdir = tempfile.mkdtemp()
    filename_shp = os.path.join(tmpdir, 'test.shp')
    
    create_sample_data(filename_shp, driver='ESRI Shapefile')
    collection = fiona.open(filename_shp, 'r')
    fiona.remove(collection)
    assert(not os.path.exists(filename_shp))

def test_remove_path_without_driver(tmpdir=None):
    if tmpdir is None:
        tmpdir = tempfile.mkdtemp()
    filename_shp = os.path.join(tmpdir, 'test.shp')

    create_sample_data(filename_shp, driver='ESRI Shapefile')

    with pytest.raises(Exception):
        fiona.remove(filename_shp)

    assert(os.path.exists(filename_shp))
